from .model import TBlock


CFG1 = [
    TBlock('Block1',
           in_channels=3,
           out_channels=32,
           spatial_ksize=3,
           spatial_stride=1,
           temporal_ksize=2,
           temporal_stride=2,
           temporal_pool_type='max',
           cg_ksize=3,
           cg_stride=2),
    TBlock('Block1',
           in_channels=32,
           out_channels=64,
           spatial_ksize=3,
           spatial_stride=1,
           temporal_ksize=3,
           temporal_stride=1,
           temporal_pool_type='avg',
           cg_ksize=3,
           cg_stride=2),
    TBlock('Block2',
           in_channels=64,
           out_channels=128,
           temporal_ksize=3,
           temporal_stride=1),
    TBlock('Block3',
           in_channels=128,
           spatial_ksize=3,
           temporal_ksize=3),
    TBlock('Block3',
           in_channels=128,
           spatial_ksize=3,
           temporal_ksize=3),
    TBlock('Block3',
           in_channels=128,
           spatial_ksize=3,
           temporal_ksize=3),
    TBlock('Block4',
           in_channels=128,
           out_channels=256,
           temporal_ksize=3,
           temporal_stride=1,
           cg_ksize=3),
    TBlock('Block4',
           in_channels=256,
           out_channels=512,
           temporal_ksize=3,
           temporal_stride=1,
           cg_ksize=3),
]
